@@ -70,7 +70,7 @@ import static org.jruby.runtime.Visibility.PRIVATE;
 public class RubyStringScanner extends RubyObject {
     private static final long serialVersionUID = -3722138049229128675L;
 
-    private static final String STRSCAN_VERSION = "3.1.2";
+    private static final String STRSCAN_VERSION = "3.1.4";
 
     private RubyString str;
     private int curr = 0;
@@ -108,6 +108,15 @@ public class RubyStringScanner extends RubyObject {
         matched = false;
     }
 
+    private void clearNamedCaptures() {
+        pattern = null;
+    }
+
+    private void clearMatchStatus() {
+        clearMatched();
+        clearNamedCaptures();
+    }
+
     private void setMatched() {
         matched = true;
     }
@@ -167,7 +176,7 @@ public class RubyStringScanner extends RubyObject {
     public IRubyObject reset(ThreadContext context) {
         check(context);
         curr = 0;
-        clearMatched();
+        clearMatchStatus();
         return this;
     }
 
@@ -175,7 +184,7 @@ public class RubyStringScanner extends RubyObject {
     public IRubyObject terminate(ThreadContext context) {
         check(context);
         curr = str.getByteList().getRealSize();
-        clearMatched();
+        clearMatchStatus();
         return this;
     }
 
@@ -198,7 +207,7 @@ public class RubyStringScanner extends RubyObject {
     public IRubyObject set_string(ThreadContext context, IRubyObject str) {
         this.str = RubyString.stringValue(str);
         curr = 0;
-        clearMatched();
+        clearMatchStatus();
         return str;
     }
 
@@ -265,7 +274,7 @@ public class RubyStringScanner extends RubyObject {
     private IRubyObject scan(ThreadContext context, IRubyObject regex, boolean succptr, boolean getstr, boolean headonly) {
         final Ruby runtime = context.runtime;
         check(context);
-        clearMatched();
+        clearMatchStatus();
 
         int restLen = restLen();
         if (restLen < 0) {
@@ -312,13 +321,13 @@ public class RubyStringScanner extends RubyObject {
                 if (ByteList.memcmp(strBL.unsafeBytes(), currPtr, patternBL.unsafeBytes(), patternBL.begin(), patternSize) != 0) {
                     return context.nil;
                 }
-                setRegisters(patternSize);
+                setRegisters(0, patternSize);
             } else {
                 int pos = StringSupport.index(strBL, patternBL, currPtr, patternEnc);
                 if (pos == -1) {
                     return context.nil;
                 }
-                setRegisters(patternSize + pos - curr);
+                setRegisters(pos - curr, patternSize);
             }
         }
 
@@ -371,11 +380,11 @@ public class RubyStringScanner extends RubyObject {
     }
 
     // MRI: set_registers
-    private void setRegisters(int length) {
+    private void setRegisters(int pos, int length) {
         if (fixedAnchor) {
-            regs = Region.newRegion(curr, curr + length);
+            regs = Region.newRegion(pos + curr, pos + curr + length);
         } else {
-            regs = Region.newRegion(0, length);
+            regs = Region.newRegion(pos, pos + length);
         }
     }
 
@@ -453,7 +462,7 @@ public class RubyStringScanner extends RubyObject {
 
     public IRubyObject getchCommon(ThreadContext context) {
         check(context);
-        clearMatched();
+        clearMatchStatus();
         ByteList strBL = str.getByteList();
         int strSize = strBL.getRealSize();
 
@@ -481,7 +490,7 @@ public class RubyStringScanner extends RubyObject {
     @JRubyMethod(name = "get_byte")
     public IRubyObject get_byte(ThreadContext context) {
         check(context);
-        clearMatched();
+        clearMatchStatus();
         if (curr >= str.getByteList().getRealSize()) return context.nil;
 
         prev = curr;
@@ -508,7 +517,7 @@ public class RubyStringScanner extends RubyObject {
     public IRubyObject scan_byte(ThreadContext context) {
         Ruby runtime = context.runtime;
         check(context);
-        clearMatched();
+        clearMatchStatus();
         ByteList byteList = str.getByteList();
         int curr = this.curr;
         if (curr >= byteList.getRealSize()) return context.nil;
@@ -562,7 +571,7 @@ public class RubyStringScanner extends RubyObject {
     public IRubyObject scan_base10_integer(ThreadContext context) {
         final Ruby runtime = context.runtime;
         check(context);
-        clearMatched();
+        clearMatchStatus();
 
         strscanMustAsciiCompat(runtime);
 
@@ -584,7 +593,6 @@ public class RubyStringScanner extends RubyObject {
             return context.nil;
         }
 
-        setMatched();
         prev = ptr;
 
         while (len < remaining_len && Character.isDigit(bytes.get(ptr + len))) {
@@ -598,7 +606,7 @@ public class RubyStringScanner extends RubyObject {
     public IRubyObject scan_base16_integer(ThreadContext context) {
         final Ruby runtime = context.runtime;
         check(context);
-        clearMatched();
+        clearMatchStatus();
 
         strscanMustAsciiCompat(runtime);
 
@@ -617,7 +625,10 @@ public class RubyStringScanner extends RubyObject {
             len++;
         }
 
-        if ((remaining_len >= (len + 2)) && bytes.get(ptr + len) == '0' && bytes.get(ptr + len + 1) == 'x') {
+        if ((remaining_len >= (len + 3)) &&
+                bytes.get(ptr + len) == '0' &&
+                bytes.get(ptr + len + 1) == 'x' &&
+                isHexChar(bytes.get(ptr + len + 2))) {
             len += 2;
         }
 
@@ -625,8 +636,6 @@ public class RubyStringScanner extends RubyObject {
             return context.nil;
         }
 
-        setMatched();
-        adjustRegisters();
         prev = ptr;
 
         while (len < remaining_len && isHexChar(bytes.get(ptr + len))) {
@@ -639,6 +648,9 @@ public class RubyStringScanner extends RubyObject {
     private RubyInteger strscanParseInteger(Ruby runtime, ByteList bytes, int ptr, int len, int base) {
         this.curr = ptr + len;
 
+        setMatched();
+        adjustRegisters();
+
         return ConvertBytes.byteListToInum(runtime, bytes, ptr, len, base, true);
     }
 
@@ -664,7 +676,7 @@ public class RubyStringScanner extends RubyObject {
         }
 
         curr = prev;
-        clearMatched();
+        clearMatchStatus();
 
         return this;
     }
@@ -739,12 +751,14 @@ public class RubyStringScanner extends RubyObject {
             return context.nil;
         }
 
-        if (idx instanceof RubySymbol || idx instanceof RubyString) {
-            if (pattern == null) return context.nil;
-        }
-
         Ruby runtime = context.runtime;
 
+        if (idx instanceof RubySymbol || idx instanceof RubyString) {
+            if (pattern == null) {
+                throw runtime.newRaiseException((RubyClass) getMetaClass().getConstant("IndexError"), "undefined group name reference: " + idx);
+            }
+        }
+
         int i = RubyMatchData.backrefNumber(runtime, pattern, regs, idx);
 
         return extractRegion(context, i);
@@ -843,7 +857,7 @@ public class RubyStringScanner extends RubyObject {
     }
 
     @JRubyMethod(name = "named_captures")
-    public IRubyObject named_captured(ThreadContext context) {
+    public IRubyObject named_captures(ThreadContext context) {
         Ruby runtime = context.runtime;
         IRubyObject nil = context.nil;
 
@@ -858,7 +872,10 @@ public class RubyStringScanner extends RubyObject {
             IRubyObject value = nil;
 
             for (int i : nameEntry.getBackRefs()) {
-                value = extractRegion(context, i);
+                IRubyObject v = extractRegion(context, i);
+                if (v != nil) {
+                    value = v;
+                }
             }
 
             int nameP = nameEntry.nameP;
